function FxDTSBrick::sendPower(%this, %n)
{
  %db = %this.getDataBlock();

  if (%db.isGen)
    %power = %db.output;
  else if (%db.isWire)
    %power = %this.power - 0.01;
  else
    %power = 0;

  if (%this.isDead() || %this.isFakeDead() || %power <= 0)
    return;

  %box = %this.getWorldBox();
  %box = vectorSub(getWords(%box, 3, 5), getWords(%box, 0, 2));
  %box = vectorScale(%box, 1.25);

  initContainerBoxSearch(
    %this.getWorldBoxCenter(),
    %box, $TypeMasks::FxBrickObjectType
  );

  while (isObject(%obj = containerSearchNext()))
  {
    if (%obj == %this)
      continue;

    if (!%obj.getDataBlock().isWire && !%obj.getDataBlock().isGate)
      continue;

    if (%obj.isDead() || %obj.isFakeDead())
      continue;

    if (!%this.isDiagonalTo(%obj) || %obj == %this.source)
      continue;

    %count++;
  }

  %power /= %count;

  initContainerBoxSearch(
    %this.getWorldBoxCenter(),
    %box, $TypeMasks::FxBrickObjectType
  );

  while (isObject(%obj = containerSearchNext()))
  {
    if (%obj == %this)
      continue;

    if (%obj.isDead() || %obj.isFakeDead())
      continue;

    if (!%this.isDiagonalTo(%obj) || %obj == %this.source)
      continue;

    if (%obj.getDataBlock().isGate)
    {
      echo("reached gate " @ %obj @ " with power " @ %power);

      if (!%obj.state)
      {
        %gates = appendWord(%gates, %obj);
        %obj.state++;
      }

      %dir = vectorNormalize(vectorSub(%obj.position, %this.position));
      %obj.powers = appendWord(%obj.powers, %power);
      echo(" +- power list: " @ %obj.powers);

      continue;
    }

    if (!%obj.getDataBlock().isWire)
      continue;

    if (%obj.power > 0)
    {
      %obj.setPower(%obj.power + %power, %this);
      continue;
    }

    %obj.setPower(%power, %this);
    %wires = appendWord(%wires, %obj);
  }

  return %wires TAB %gates;
}

function FxDTSBrick::processGate(%this)
{
  if (%this.state >= 2)
    return;

  %this.state++;
  echo("processing gate " @ %this);
  echo(" +- depth: " @ %this.state - 1);

  %db = %this.getDataBlock();
  %power = 0;

  echo(" +- power list: " @ %this.powers);

  if (%this.powers !$= "")
  {
    %count = getWordCount(%this.powers);

    for (%i = 0; %i < %count; %i++)
      %power += getWord(%this.powers, %i);
  }

  %logic = %db.processGate(%this);
  echo(%power SPC "*" SPC %logic SPC "=>" SPC %power * %logic);
  %power *= mClampF(%logic, 0, 1);

  echo("output = " @ %power);
  %this.output = %power;

  // if (%power <= 0)
  //   return "";

  %ang = (%this.angleId + %db.orientationFix) % 4;

  switch (%ang)
  {
    case 0: %dir = "1 0";
    case 1: %dir = "0 -1";
    case 2: %dir = "-1 0";
    case 3: %dir = "0 1";
  }

  %pos = %this.getRayOrigin();
  %dir = vectorScale(%dir, 0.5);

  %ray = containerRayCast(
    %pos, vectorAdd(%pos, %dir),
    $TypeMasks::FxBrickObjectType, %this
  );

  // %proj = new Projectile()
  // {
  //   datablock = GunProjectile;
  //   scale = "0.2 0.2 0.2";
  //   initialPosition = vectorAdd(%pos, "0 0 0.25");
  //   initialVelocity = %dir;
  // };

  // MissionCleanup.add(%proj);

  if (!%ray)
    return;

  %col = getWord(%ray, 0);

  if (%col.isDead() || %col.isFakeDead())
    return "";

  if (!%col.isDiagonalTo(%this))
    return "";

  // if (%col.getDataBlock().isGate)
  // {
  //   %dir = vectorNormalize(vectorSub(%col.position, %this.position));
  //   %obj.powers = appendWord(%col.powers, %power);

  //   if (!%col.state)
  //   {
  //     %col.state++;
  //     //return "" TAB %col;
  //   }

  //   return "";
  // }

  if (!%col.getDataBlock().isWire || %col.power > 0)
    return "";

  %col.setPower(%power, %obj);
  return %col;
}

function FxDTSBrick::setPower(%this, %power, %source)
{
  //%this.schedule($n++*1000,flash);

  %this.power = getMax(%power, 0);
  %this.source = %source;

  //%this.setColorFX(%this.power <= 0 ? 0 : 3);
  cancel(%this.colorFX);

  if (%source)
    %this.colorFX = %this.schedule($n++ * 50, setColorFX, %this.power <= 0 ? 0 : 3);
  else
    %this.setColorFX(%this.power <= 0 ? 0 : 3);

  if (isObject(%this.shape))
    %this.shape.setShapeName(%this.power <= 0 ? "" : %this.power);
}

function updatePower()
{
  $n = 0;
  %rate = 0.2;

  if ($Sim::Time - $LastPowerUpdate < %rate)
  {
    %time = (%rate - ($Sim::Time - $LastPowerUpdate)) * 1000;

    if (!isEventPending($PowerUpdateSchedule))
      $PowerUpdateSchedule = schedule(%time, 0, updatePower);

    return;
  }

  cancel($PowerUpdateSchedule);
  $LastPowerUpdate = $Sim::Time;

  // Step 1: Clear all power.
  %count = WireGroup.getCount();

  for (%i = 0; %i < %count; %i++)
    WireGroup.getObject(%i).setPower(0, 0);

  %count = GateGroup.getCount();

  for (%i = 0; %i < %count; %i++)
  {
    %gate = GateGroup.getObject(%i);

    %gate.state = 0;
    %gate.powers = "";
  }

  // Step 2: Populate queue with generators.
  %count = GenGroup.getCount();

  for (%i = 0; %i < %count; %i++)
    %pendingWires = appendWord(%pendingWires, GenGroup.getObject(%i));

  // Step 3: While the queue isn't empty..
  while (%pendingWires !$= "")
  {
    // .. empty it
    while (%pendingWires !$= "")
    {
      %obj = getWord(%pendingWires, 0);
      %pendingWires = removeWord(%pendingWires, 0);

      if (!isObject(%obj))
        continue;

      %temp = %obj.sendPower(%n);

      %pendingWires = appendWord(%pendingWires, getField(%temp, 0));
      %pendingGates = appendWord(%pendingGates, getField(%temp, 1));
    }

    // .. and *try* to fill it again
    if (%pendingGates $= "")
      continue;

    while (%pendingGates !$= "" && %shh++ <= 1000)
    {
      %gate = getWord(%pendingGates, 0);
      %pendingGates = removeWord(%pendingGates, 0);

      if (!isObject(%obj))
        continue;

      %temp = %gate.processGate();

      %pendingWires = appendWord(%pendingWires, getField(%temp, 0));
      %pendingGates = appendWord(%pendingGates, getField(%temp, 1));
    }
  }
}
